# -*- coding: utf-8 -*-
"""
Created on Thu Nov  5 10:31:51 2020

@author: mathiash
"""
import concurrent
import cantera as ct
import numpy as np
import lmfit
import os

def calculate_gulder_coefficient(inputs):
    "declare variables"
    gulder={}
    LBV={}
    "get phi data"
    refT = 300 #reference temperature for gülder coeficient use in OpenFOAM, see thermophysicalModels/laminarFlameSpeed/Gulders/Gulders.C
    refP = 1.013e+5 #reference pressure for gülder coeficient use in OpenFOAM, see thermophysicalModels/laminarFlameSpeed/Gulders/Gulders.C
    refPhi = 1.0 #reference phi for reference laminar burning velocty gülder coeficient
    dPhi=0.5; dT=50; dP=0.5e+5
    phiList=np.arange(0.5,1.5+dPhi,dPhi) #list of equivalence ratio used for gulder coefficient 
    oldPath = os.getcwd() # keep path
    os.chdir(inputs['reaction mech path']) #set path to cantera mechansims
    gulder['omega'], gulder['eta'], gulder['xi'], LBV['phi'] = fit_gulder_coefficients_to_phi_and_LBV(inputs,refT,refP,phiList)
    import auxiliaryFunctions as af
    posPhiOne = af.find_index_closest_to_value(phiList, refPhi) #Find position and value of Sl(phi=1)
    refSu=LBV['phi'][posPhiOne] # # reference laminar burning velocty in gülder coeficient
    tempList = np.arange(300,400+dT,dT) #list of temperatures used for gulder coefficient 
    gulder['alpha'], LBV['temp']= fit_gulder_coefficients_to_temperature_and_LBV(inputs,tempList,refT,refP,refPhi,refSu)
    pList = np.arange(0.513e+5,2.013e+5+dP,dP) #list of pressure used for gulder coefficient 
    gulder['beta'], LBV['pressure']= fit_gulder_coefficients_to_pressure_and_LBV(inputs,pList,refT,refP,refPhi,refSu)
    os.chdir(oldPath)
    return gulder, LBV

def run_FreeFlame(mech,fuel,air,T,p,phi):
    gas = ct.Solution(mech)
    gas.set_equivalence_ratio(phi,fuel,air)
    gas.TP = T, p
    f=ct.FreeFlame(gas,width=0.4)
    f.set_max_time_step(1000)
    f.set_refine_criteria(ratio=3,slope=0.06,curve=0.12)
    f.transport_model='Multi'
    f.solve(0,auto=True,refine_grid=True)
    return phi,f.u[0],f.u[-1]

def calcualte_LBV_by_multiprocessors(mech,fuel,oxidant,T,p,phi):
    output = list()
    if np.size(phi) > 1:
        with concurrent.futures.ProcessPoolExecutor(max_workers=3) as executor:
            results = ([executor.submit(run_FreeFlame,mech,fuel,oxidant,T,p,i) for i in phi])
            for f in concurrent.futures.as_completed(results):
                output.append(f.result())
    elif np.size(T) > 1:
        with concurrent.futures.ProcessPoolExecutor(max_workers=3) as executor:
            results = [executor.submit(run_FreeFlame,mech,fuel,oxidant,i,p,phi) for i in T]
            for f in concurrent.futures.as_completed(results):
                output.append(f.result())
    elif np.size(p) > 1:
        with concurrent.futures.ProcessPoolExecutor(max_workers=3) as executor:
            results = [executor.submit(run_FreeFlame,mech,fuel,oxidant,T,i,phi) for i in p]
            for f in concurrent.futures.as_completed(results):
                output.append(f.result())
    sortedResults = sorted(output,key=lambda x:x[0])
    Su=[i[1] for i in sortedResults]; Su=np.array(Su)
    Sb=[i[2] for i in sortedResults]; Sb=np.array(Sb) 
    return Su, Sb

def fit_gulder_coefficients_to_phi_and_LBV(inputs,referenceTemp,referencePressure,phiRange):
    Su,Sb = calcualte_LBV_by_multiprocessors(inputs['reaction mech'],inputs['fuel'],inputs['oxidant'],referenceTemp,referencePressure,phiRange)
    pars = lmfit.Parameters()
    pars.add_many(('omega', 0.5 ), ('eta', 0.1), ('xi',5))
    def fit_gulder_coefficients_to_phi_range(p):
        return Su - (p['omega']*phiRange**(p['eta'])*np.exp(-p['xi']*(phiRange-1.075)**2))
    mini = lmfit.Minimizer(fit_gulder_coefficients_to_phi_range, pars)
    result = mini.minimize()
    return result.params['omega'].value, result.params['eta'].value, result.params['xi'].value, Su

def fit_gulder_coefficients_to_temperature_and_LBV(inputs,tempRange,referenceTemp,referencePressure,referencePhi,referenceSu):
    Su, Sb = calcualte_LBV_by_multiprocessors(inputs['reaction mech'],inputs['fuel'],inputs['oxidant'],tempRange,referencePressure, referencePhi)
    pars = lmfit.Parameters()
    pars.add_many(('alpha', 1.75 ))
    def fit_gulder_coefficient_to_temperature_range(p):
        return Su - referenceSu*(tempRange/referenceTemp)**p['alpha']
    mini = lmfit.Minimizer(fit_gulder_coefficient_to_temperature_range, pars)
    result = mini.minimize()
    return result.params['alpha'].value, Su

def fit_gulder_coefficients_to_pressure_and_LBV(inputs,pressureRange,referenceTemp,referencePressure,referencePhi,referenceSu):
    Su, Sb = calcualte_LBV_by_multiprocessors(inputs['reaction mech'],inputs['fuel'],inputs['oxidant'],referenceTemp,pressureRange,referencePhi)
    pars = lmfit.Parameters()
    pars.add_many(('beta', -0.3))
    def fit_gulder_coefficient_to_pressure_range(p):
        return Su - referenceSu*(pressureRange/referencePressure)**p['beta']
    mini = lmfit.Minimizer(fit_gulder_coefficient_to_pressure_range, pars)
    result = mini.minimize()
    return result.params['beta'].value, Su

#import auxiliaryFunctions as af
#inputers={}
#inputers['reaction mech'], inputers['reaction mech path'] = af.get_the_filename_and_path_of_the_reaction_mechanisms(3) # Reaction mechanisms filename
#inputers['fuel']='h2:0.35,ch4:0.65'; 
#inputers['oxidant']='o2:1,n2:'+str(79/21)+''
#g, sl = calculate_gulder_coefficient(inputers)

#if __name__ == '__main__':
#    import auxiliaryFunctions as af
#    inputers={}
#    inputers['reaction mech'], inputers['reaction mech path'] = af.get_the_filename_and_path_of_the_reaction_mechanisms(3) # Reaction mechanisms filename
#    inputers['fuel']='h2:0.35,ch4:0.65'; 
#    inputers['oxidant']='o2:1,n2:'+str(79/21)+''
#    g, sl = calculate_gulder_coefficient(inputers)
    