# -*- coding: utf-8 -*-
"""
Code in article
@author: Mathias Heniksen 
doi: article
"""

import cantera as ct
import numpy as np
import os
import lmfit
import auxiliaryFunctions as af

def calculate_NASApolynomial_coefficients_for_mixture(gas,species,T,p):
    "declare variables"
    comTlist=list(); minTlist=list()
    maxTlist=list(); fractions=list()
    NASAlowPoly=0 ;  NASAhighPoly=0
    "set gas initial conditions for species"
    listOfSpecies, listOfFraction = af.find_name_and_fraction_of_species_from_cantera_input_string(species)
    gas.X =species
    gas.TP = T,p
    "Find and calculate polynomial coefficients based on specie fraction"
    for i, specie in enumerate(listOfSpecies):
        T_min, T, T_max, NASAhigh, NASAlow= find_NASApolynomial_coefficient_for_specie(gas,specie)
        comTlist.append(T)
        minTlist.append(T_min)
        maxTlist.append(T_max)
        idx = gas.species_index(specie)
        if i > 0:
            if gas.X[idx] > max(fractions):
                fractions.append(gas.X[idx])
                commonT =T
                minT=T_min
                maxT=T_max
        else:
            fractions.append(gas.X[idx])
            commonT =T
            minT=T_min
            maxT=T_max
        NASAlowPoly += gas.X[idx]*NASAlow
        NASAhighPoly += gas.X[idx]*NASAhigh
    report =  generate_message_report_for_NASApolynomial_in_thermoReport(minTlist,comTlist,maxTlist)
    outputData={}#generate a common output data file
    outputData['common ref temp']= commonT; outputData['minimum ref temp']= minT;
    outputData['maximum ref temp']= maxT; outputData['NASApoly low Coefficients']=NASAlowPoly;
    outputData['NASApoly high Coefficients']=NASAhighPoly
    return outputData, report

def generate_message_report_for_NASApolynomial_in_thermoReport(minimumTempList, commenTempList, maximumTempList):
    report={}
    report['minimum ref temp']= generate_validation_message_from_refernce_temperature_list(minimumTempList)
    report['common ref temp']= generate_validation_message_from_refernce_temperature_list(commenTempList)
    report['maximum ref temp']= generate_validation_message_from_refernce_temperature_list(maximumTempList)
    return report

def generate_validation_message_from_refernce_temperature_list(tempertureList):
    if(len(set(tempertureList))==1):
        msg="The temperatures are equal for all species"
    else:
        T_max=max(tempertureList)
        T_min=min(tempertureList)
        msg='The temperature are not equal for all species. The temperature range that is not shared is '+str(T_min)+' - '+str(T_max)
    return msg

def find_NASApolynomial_coefficient_for_specie(gas,specie):
    idx = gas.species_index(specie)
    data = gas.species()[idx].thermo.coeffs
    T_low= gas.species()[idx].thermo.min_temp
    T_max=gas.species()[idx].thermo.max_temp
    T_com= data[0]
    coef_high = data[1:8]
    coef_low = data[8:]
    return T_low, T_com, T_max, coef_high, coef_low

def calculate_burnt_product_gas_composition(gas,fuel,oxidant,T,p):
    "declare variables"
    combinedList = list()
    listOfFraction=list()
    listOfSpecies=list()
    burntProducts=''
    gas.set_equivalence_ratio(1,fuel,oxidant)
    gas.TP = T, p
    gas.equilibrate('HP')
    for specie in gas.species_names:
        index = gas.species_index(specie)
        if gas.X[index] >= 0.001:
            listOfSpecies.append(specie)
            listOfFraction.append(gas.X[index])
            
    listOfFraction = listOfFraction/sum(listOfFraction)
    for specie, fraction in zip(listOfSpecies,listOfFraction):
        combinedList.append(specie+':'+str(fraction)+',')
    burntProducts=burntProducts.join(combinedList)
    burntProducts=burntProducts[:-1]
    return burntProducts

def calculate_sutherland_coefficients_for_gas_mixture(gas, species, minT, maxT, p):
    gas.transport_model = 'Multi'
    gas.X=species
    rangeT=np.arange(minT, maxT,1)
    mu=list()
    for T in rangeT:
        gas.TP = T, p
        mu.append(gas.viscosity)
    pars = lmfit.Parameters()
    pars.add_many(('As', 1.67e-06 ), ('Ts', 170))
    def fitter(p):
        return mu - (p['As']*np.sqrt(rangeT))/(1+p['Ts']/rangeT)
    mini = lmfit.Minimizer(fitter, pars)
    result = mini.minimize()
    #print(lmfit.fit_report(result.params))
    return result.params['As'].value, result.params['Ts'].value 

def generate_thermoProperties_data_for_3_species_combustion(inputParameters):
    "declare variables"
    oldPath = os.getcwd()
    os.chdir(inputParameters['reaction mech path'])
    gas = ct.Solution(inputParameters['reactionMech'])
    os.chdir(oldPath)
    report={}
    "Calculate stoichiometric AirFuel mass ratio for input parameters"
    inputParameters['Stoich AirFuel mass ratio'],inputParameters['Stoich Fuel mass fraction'] = af.calculate_air_fuel_mass_ratio(gas,inputParameters['fuel'],
                                                                                                                  inputParameters['oxidant'],
                                                                                                                  inputParameters['T0'],
                                                                                                                  inputParameters['p0'])    
    "Calcualte burnt product gas composition for input parameters"
    inputParameters['burntProducts'] = calculate_burnt_product_gas_composition(gas,inputParameters['fuel'],
                                                                                      inputParameters['oxidant'],
                                                                                      inputParameters['T0'],
                                                                                      inputParameters['p0'])
    
    "oxidant output parameters"
    oxidantThermoData, report['oxidant'] = calculate_NASApolynomial_coefficients_for_mixture(gas,inputParameters['oxidant'],
                                                                                      inputParameters['T0'],
                                                                                      inputParameters['p0'])
    oxidantThermoData['mole weight']=af.get_mole_weight_for_gas_mixture(gas,inputParameters['oxidant'],inputParameters['T0'], 
                                                                     inputParameters['p0'])
    oxidantThermoData['As'], oxidantThermoData['Ts'] = calculate_sutherland_coefficients_for_gas_mixture(gas,inputParameters['oxidant'],
                                                                                                         oxidantThermoData['minimum ref temp'],
                                                                                                         oxidantThermoData['maximum ref temp'],
                                                                                                         inputParameters['p0'])
    "Fuel output parameters"
    fuelThermoData, report['fuel'] = calculate_NASApolynomial_coefficients_for_mixture(gas,inputParameters['fuel'],
                                                                                      inputParameters['T0'],
                                                                                      inputParameters['p0'])
    fuelThermoData['mole weight']=af.get_mole_weight_for_gas_mixture(gas,inputParameters['fuel'],inputParameters['T0'], 
                                                                     inputParameters['p0'])
    fuelThermoData['As'], fuelThermoData['Ts'] = calculate_sutherland_coefficients_for_gas_mixture(gas,inputParameters['fuel'],
                                                                                                         fuelThermoData['minimum ref temp'],
                                                                                                         fuelThermoData['maximum ref temp'],
                                                                                                         inputParameters['p0'])

    "Burnt Product output parameters"
    burntProductsThermoData, report['burntProducts'] = calculate_NASApolynomial_coefficients_for_mixture(gas,inputParameters['burntProducts'],
                                                                                      inputParameters['T0'],
                                                                                      inputParameters['p0'])
    burntProductsThermoData['mole weight']=af.get_mole_weight_for_gas_mixture(gas,inputParameters['burntProducts'],inputParameters['T0'], 
                                                                     inputParameters['p0'])
    burntProductsThermoData['As'], burntProductsThermoData['Ts'] = calculate_sutherland_coefficients_for_gas_mixture(gas,inputParameters['burntProducts'],
                                                                                                         burntProductsThermoData['minimum ref temp'],
                                                                                                         burntProductsThermoData['maximum ref temp'],
                                                                                                         inputParameters['p0'])
    return inputParameters, oxidantThermoData, fuelThermoData, burntProductsThermoData, report
