# -*- coding: utf-8 -*-
"""
@author: Mathias Henriksen
Doi
"""

import os
import numpy as np
import matplotlib.pyplot as plt
"""
-------------------------------------------------------------------------------
This section contains all function that are related to the Gulder Coefficients 
calcualtion except the run and input  function. For windows systems, the author
of the code has been unable to get multiprocessing to work across scripts. The 
solution was to keep all scripts in the main script. It is also important that 
this script is run as __main__ for multiprocessing to work
"""
import concurrent
import cantera as ct
import lmfit
import pandas as pd

def calculate_gulder_coefficient(inputs):
    "declare variables"
    gulder={}
    LBV={}
    "get phi data"
    refT = 300.0 #reference temperature for gülder coeficient use in OpenFOAM, see thermophysicalModels/laminarFlameSpeed/Gulders/Gulders.C
    refP = 1.013e+5 #reference pressure for gülder coeficient use in OpenFOAM, see thermophysicalModels/laminarFlameSpeed/Gulders/Gulders.C
    refPhi = 1.0 #reference phi for reference laminar burning velocty gülder coeficient
    dPhi=0.05; dT=25; dP=0.5e+5
    phiList=np.arange(0.4,1.8+dPhi,dPhi,float) #list of equivalence ratio used for gulder coefficient 
    LBV['phi list']=phiList
    oldPath = os.getcwd() # keep path
    os.chdir(inputs['reaction mech path']) #set path to cantera mechansims
    gulder['omega'], gulder['eta'], gulder['xi'], LBV['phi'] = fit_gulder_coefficients_to_phi_and_LBV(inputs,refT,refP,phiList)
    print('phi-calculation done')
    import auxiliaryFunctions as af
    posPhiOne = af.find_index_closest_to_value(phiList, refPhi) #Find position and value of Sl(phi=1)
    refSu=LBV['phi'][posPhiOne] # # reference laminar burning velocty in gülder coeficient
    tempList = np.arange(250,700+dT,dT,float) #list of temperatures used for gulder coefficient 
    LBV['temp list']=tempList
    gulder['alpha'], LBV['temp']= fit_gulder_coefficients_to_temperature_and_LBV(inputs,tempList,refT,refP,refPhi,refSu)
    print('temp-calculation done')
    pList = np.arange(0.513e+5,10.013e+5+dP,dP,float) #list of pressure used for gulder coefficient 
    LBV['pressure list']=pList
    gulder['beta'], LBV['pressure']= fit_gulder_coefficients_to_pressure_and_LBV(inputs,pList,refT,refP,refPhi,refSu)
    os.chdir(oldPath)
    inputs['reference temperature'] = refT; inputs['reference pressure'] = refP
    inputs['reference LBV at phi 1.0'] = refSu
    inputs['equivalence ratio range for fit'] = str(min(phiList))+' -  '+str(max(phiList))+' - interval = '+str(dPhi) 
    inputs['temperatue range for fit'] = str(min(tempList))+' - '+str(max(tempList))+' - interval = '+str(dT) 
    inputs['pressure range for fit'] = str(min(pList))+' - '+str(max(pList))+' - interval = '+str(dP)
    return gulder, LBV, inputs

def run_FreeFlame(mech,fuel,air,T,p,phi,listOf):
    gas = ct.Solution(mech)
    gas.set_equivalence_ratio(phi,fuel,air)
    gas.TP = T, p
    f=ct.FreeFlame(gas,width=0.4)
    f.set_max_time_step(1000)
    f.set_refine_criteria(ratio=3,slope=0.06,curve=0.12)
    f.transport_model='Multi'
    f.solve(0,auto=True,refine_grid=True)
    if listOf == 'phi':
        return phi,f.u[0],f.u[-1]
    elif listOf == 'temperature':
        return T,f.u[0],f.u[-1]
    elif listOf == 'pressure':
        return p,f.u[0],f.u[-1]

def calcualte_LBV_by_multiprocessors(mech,fuel,oxidant,T,p,phi):
    output = list()
    if np.size(phi) > 1:
        with concurrent.futures.ProcessPoolExecutor(max_workers=4) as executor:
            results = ([executor.submit(run_FreeFlame,mech,fuel,oxidant,T,p,i,'phi') for i in phi])
            for f in concurrent.futures.as_completed(results):
                output.append(f.result())
    elif np.size(T) > 1:
        with concurrent.futures.ProcessPoolExecutor(max_workers=4) as executor:
            results = [executor.submit(run_FreeFlame,mech,fuel,oxidant,i,p,phi,'temperature') for i in T]
            for f in concurrent.futures.as_completed(results):
                output.append(f.result())
    elif np.size(p) > 1:
        with concurrent.futures.ProcessPoolExecutor(max_workers=4) as executor:
            results = [executor.submit(run_FreeFlame,mech,fuel,oxidant,T,i,phi,'pressure') for i in p]
            for f in concurrent.futures.as_completed(results):
                output.append(f.result())
    sortedResults = sorted(output,key=lambda x:x[0])
    Su=[i[1] for i in sortedResults]; Su=np.array(Su)
    Sb=[i[2] for i in sortedResults]; Sb=np.array(Sb) 
    return Su, Sb

def fit_gulder_coefficients_to_phi_and_LBV(inputs,referenceTemp,referencePressure,phiRange):
    Su,Sb = calcualte_LBV_by_multiprocessors(inputs['reaction mech'],inputs['fuel'],inputs['oxidant'],referenceTemp,referencePressure,phiRange)
    pars = lmfit.Parameters()
    pars.add_many(('omega', 0.5 ), ('eta', 0.1), ('xi',5))
    def fit_gulder_coefficients_to_phi_range(p):
        return Su - (p['omega']*phiRange**(p['eta'])*np.exp(-p['xi']*(phiRange-1.075)**2))
    mini = lmfit.Minimizer(fit_gulder_coefficients_to_phi_range, pars)
    result = mini.minimize()
    return result.params['omega'].value, result.params['eta'].value, result.params['xi'].value, Su

def fit_gulder_coefficients_to_temperature_and_LBV(inputs,tempRange,referenceTemp,referencePressure,referencePhi,referenceSu):
    Su, Sb = calcualte_LBV_by_multiprocessors(inputs['reaction mech'],inputs['fuel'],inputs['oxidant'],tempRange,referencePressure, referencePhi)
    pars = lmfit.Parameters()
    pars.add_many(('alpha', 1.75 ))
    def fit_gulder_coefficient_to_temperature_range(p):
        return Su - referenceSu*(tempRange/referenceTemp)**p['alpha']
    mini = lmfit.Minimizer(fit_gulder_coefficient_to_temperature_range, pars)
    result = mini.minimize()
    return result.params['alpha'].value, Su

def fit_gulder_coefficients_to_pressure_and_LBV(inputs,pressureRange,referenceTemp,referencePressure,referencePhi,referenceSu):
    Su, Sb = calcualte_LBV_by_multiprocessors(inputs['reaction mech'],inputs['fuel'],inputs['oxidant'],referenceTemp,pressureRange,referencePhi)
    pars = lmfit.Parameters()
    pars.add_many(('beta', -0.3))
    def fit_gulder_coefficient_to_pressure_range(p):
        return Su - referenceSu*(pressureRange/referencePressure)**p['beta']
    mini = lmfit.Minimizer(fit_gulder_coefficient_to_pressure_range, pars)
    result = mini.minimize()
    return result.params['beta'].value, Su
"""
The section containg function for the Gulder Coefficients calcualtion end here
-------------------------------------------------------------------------------
"""

def test_gulder_coefficients():
    mainPath=os.getcwd() #keep main path
    os.chdir(mainPath+'\\subRoutines\\')
    import auxiliaryFunctions as af
#    os.chdir(path)
#    path='C:\\Users\\mathiash\\Documents\\Publications\\Mine\\Ongoing Articles\\OpenFoam simulation of Gas explosions, from reaction mechanism to XiFoam input data\\pythonScripts\\35H2_65CH4 data'
    mech, mechanismPath = af.get_the_filename_and_path_of_the_reaction_mechanisms(3)
    os.chdir(mechanismPath)
#    T0=300; #Temperature 
    p0 = 101325 #pressure
    fuel='H2:0.349,CO:0.10,CO2:0.30,CH4:0.201,C2H4:0.05' #Cantera string fuel composition in volume fractions
    oxidant='O2:1,N2:'+str(79/21)#Cantera string air/oxidant composition in volume fractions
#    os.chdir(path)
    phi = np.arange(0.5,1.6,0.1)
    Temp = np.arange(300, 700, 100)
    LBV={}
    print(Temp)
    for T in Temp:
        Su, Sb  = calcualte_LBV_by_multiprocessors(mech,fuel,oxidant,T,p0,phi)
        name = str("%.0f" %T)
        LBV[name] = Su
    
    W = 0.503872589708;  eta=-0.145517174171; Xi=4.52313414315;
    A = 1.87347293753; B = -0.37383060658;
    Su = lambda phi,T : W*phi**(eta)*np.exp(-Xi*(phi-1.075)**2)*(T/300)**A*(p0/101300)**B
    
#    SStot = np.sum((Su0 - (np.sum(Su0)/len(Su0)))**2)
#    SSres = np.sum((Su0 - canteraSu0[name])**2)

    r2={}
    for T in Temp:
        name = str("%.0f" %T)
        lbvs = Su(phi,T)
        SStot=np.sum((lbvs - np.mean(lbvs))**2)
        SSres = np.sum((LBV[name]-lbvs)**2)
        r2[name] = 1 - (SSres/SStot)
        plt.plot(phi,LBV[name],'rx')
        plt.plot(phi,lbvs,'ob')
    
#    dfLBV = pd.read_csv('Gulders_35H2_65CH4_sandiego_mech_2020_11_09.csv', sep=",")
    return r2
    
    

def test_nasa_coefficients(filename):
    mainPath=os.getcwd() #keep main path
    os.chdir(mainPath+'\\subRoutines\\')
    import calculateThermoProperties as calcTP   
    os.chdir(mainPath)
    inputs, mechanismsPath  = set_input_parameters_for_thermoPropeties_file(filename)
    inputs, oxidant, fuel, burntProducts, polynomialReport = calcTP.generate_thermoProperties_data_for_3_species_combustion(inputs)
    T=inputs['T0']
    cpMole = 8314*(oxidant['NASApoly low Coefficients'][0]+oxidant['NASApoly low Coefficients'][1]*T+oxidant['NASApoly low Coefficients'][2]*T**2+
                   oxidant['NASApoly low Coefficients'][3]*T**3+oxidant['NASApoly low Coefficients'][4]*T**4)
    cpMass = cpMole/(oxidant['mole weight'])
    import cantera as ct
    os.chdir(inputs['reaction mech path'])
    gas=ct.Solution(inputs['reactionMech'])
    gas.TPX = inputs['T0'],inputs['p0'],inputs['oxidant']
    print(gas())
    print(gas.viscosity)
    print('Cp-mass relative deviance compared with Cantera object :',np.abs(gas.cp-cpMass)/gas.cp)
    print('Cp-mole relative deviance compared with Cantera object :',np.abs(gas.cp_mole-cpMole)/gas.cp)
    return inputs

def runAll(filename):
    mainPath=os.getcwd()
    run_thermoProperties(filename)
    os.chdir(mainPath)
    run_gulder_coefficients(filename)
    return

def run_thermoProperties(filename):
    """
    runs thermoProperties generation of cantera2Foam with input the set in
    'set_input_parameters_for_thermoPropeties_file'
    """
    "import subRoutines"
    mainPath=os.getcwd() #keep main path
    os.chdir(mainPath+'\\subRoutines\\')
    import calculateThermoProperties as calcTP
    import writeThermoProperties as writeTP
    os.chdir(mainPath)
    savePath=mainPath+'\\'+filename+' data'
    if not os.path.isdir(savePath):
        os.makedirs(savePath)
    inputs, mechanismsPath  = set_input_parameters_for_thermoPropeties_file(filename)
    inputs, oxidant, fuel, burntProducts, polynomialReport = calcTP.generate_thermoProperties_data_for_3_species_combustion(inputs)
    writeTP.write_thermoProperties_file_for_Xioam_for_inhomogenous_combustion(savePath,inputs,oxidant,fuel,burntProducts,polynomialReport)
    return

def run_gulder_coefficients(nameOfCombustionMixture):
    """
    runs gulder coefficient generation part of cantera2Foam.py with input the set in
    'set_input_parameters_for_gulder_coefficient_file'
    
    nameOfCombustionMixture - a generic name for the combustion mixture 
    
    """
    "import subRoutines"
    mainPath=os.getcwd() #keep main path
    os.chdir(mainPath+'\\subRoutines\\')
    import writeSupportDocumentations as writeSD
    inputs = set_input_parameters_for_gulder_coefficient_file(nameOfCombustionMixture)
    gulderCoef,LBV, inputs = calculate_gulder_coefficient(inputs)
    os.chdir(mainPath)
    savePath=mainPath+'\\'+nameOfCombustionMixture+' data'
    if not os.path.isdir(savePath):
        os.makedirs(savePath)
    writeSD.write_gulder_output_document_and_LBV_csv(savePath,inputs,gulderCoef,LBV)
    return

def set_input_parameters_for_thermoPropeties_file(nameOfCombustionMixture='no name given'):
    """
    this function sets the input parameters for generating the thermodynmic properties.
    
    nameOfCombustionMixture - a generic name for the combustion mixture 
    """
    "check if name for mixture is given"
    if nameOfCombustionMixture=='no name given':
        nameOfCombustionMixture = input('enter a a generic name for the combustion mixture :')
    "Insert input values below"
    T0=300; #Temperature 
    p0 = 101325 #pressure
    fuel='H2:1'#Cantera string fuel composition in volume fractions
    oxidant='O2:1,N2:'+str(79/21)#Cantera string air/oxidant composition in volume fractions
    import auxiliaryFunctions as af
    mech, mechanismPath = af.get_the_filename_and_path_of_the_reaction_mechanisms(2) # Reaction mechanisms filename
    "No need to alter this function beyond this point"
    inputsThermo={} #Create common input dict as output of function
    inputsThermo['fuel']=fuel; inputsThermo['oxidant']=oxidant; 
    inputsThermo['name of combustion mixture']=nameOfCombustionMixture;
    inputsThermo['reactionMech']=mech; inputsThermo['reaction mech path']=mechanismPath
    inputsThermo['T0']=T0; inputsThermo['p0']=p0 
    return inputsThermo, mechanismPath

def set_input_parameters_for_gulder_coefficient_file(nameOfCombustionMixture):
    """
    this function sets the input parameters for generating the gülder coefficients.

    nameOfCombustionMixture - a generic name for the combustion mixture 
    """
    "insert input values below"
    fuel='H2:1' #Cantera string fuel composition in volume fractions
    oxidant='O2:1,N2:'+str(79/21)#Cantera string air/oxidant composition in volume fractions
    import auxiliaryFunctions as af
    reactionMech, mechPath = af.get_the_filename_and_path_of_the_reaction_mechanisms(2) # Reaction mechanisms filename
    "No need to alter this function beyond this point"
    inputsGulder={} #Create common input dict as output of function
    inputsGulder['fuel']=fuel; inputsGulder['oxidant']=oxidant;
    inputsGulder['name of combustion mixture']=nameOfCombustionMixture;
    inputsGulder['reaction mech']=reactionMech; inputsGulder['reaction mech path']=mechPath   
    return inputsGulder

if __name__ == '__main__':
    print('good day Milord')
    mixtureName='hydrogen'
#    r2 = test_gulder_coefficients()
    runAll(mixtureName)
#    output = test_nasa_coefficients(mixtureName)
#    run_thermoProperties(mixtureName)
#    run_gulder_coefficients(mixtureName)